# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from gitlab import GitlabGetError
from requre.online_replacing import record_requests_for_all_methods

from ogr import GitlabService
from ogr.exceptions import GitlabAPIException
from ogr.services.gitlab.pull_request import GitlabPullRequest
from tests.integration.gitlab.base import GitlabTests


@record_requests_for_all_methods()
class Service(GitlabTests):
    def test_project_create(self):
        """
        Remove https://gitlab.com/$USERNAME/new-ogr-testing-repo before data regeneration
        """
        name_of_the_repo = "new-ogr-testing-repo"
        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=self.service.user.get_username(),
        )
        with pytest.raises(GitlabGetError):
            assert project.gitlab_repo

        new_project = self.service.project_create(name_of_the_repo)
        assert new_project.repo == name_of_the_repo
        assert new_project.gitlab_repo

        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=self.service.user.get_username(),
        )
        assert project.gitlab_repo

    def test_project_create_with_description(self):
        """
        Remove the following project before data regeneration:
        https://gitlab.com/$USERNAME/new-ogr-testing-repo-with-description
        """
        name_of_the_repo = "new-ogr-testing-repo-with-description"
        description = "The description of the newly created project."

        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=self.service.user.get_username(),
        )
        with pytest.raises(GitlabGetError):
            assert project.gitlab_repo

        new_project = self.service.project_create(
            name_of_the_repo,
            description=description,
        )
        assert new_project.repo == name_of_the_repo
        assert new_project.gitlab_repo
        assert new_project.get_description() == description

        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=self.service.user.get_username(),
        )
        assert project.gitlab_repo
        assert project.get_description() == description

    def test_project_create_in_the_group(self):
        """
        Remove https://gitlab.com/packit-service/new-ogr-testing-repo-in-the-group
        before data regeneration.
        """
        name_of_the_repo = "new-ogr-testing-repo-in-the-group"
        namespace_of_the_repo = "packit-service"
        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=namespace_of_the_repo,
        )
        with pytest.raises(GitlabGetError):
            assert project.gitlab_repo

        new_project = self.service.project_create(
            repo=name_of_the_repo,
            namespace=namespace_of_the_repo,
        )
        assert new_project.repo == name_of_the_repo
        assert new_project.namespace == namespace_of_the_repo
        assert new_project.gitlab_repo

        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=namespace_of_the_repo,
        )
        assert project.gitlab_repo

    def test_project_create_duplicate(self):
        """
        Remove https://gitlab.com/$USERNAME/new-ogr-testing-repo-fail before data regeneration
        """
        name_of_the_repo = "new-ogr-testing-repo-fail"
        project = self.service.get_project(
            repo=name_of_the_repo,
            namespace=self.service.user.get_username(),
        )
        with pytest.raises(GitlabGetError):
            assert project.gitlab_repo

        self.service.project_create(name_of_the_repo)
        with pytest.raises(GitlabAPIException):
            self.service.project_create(name_of_the_repo)

    def test_service_without_auth(self):
        service = GitlabService()
        assert service.gitlab_instance
        assert service.get_project(namespace="gnachman", repo="iterm2").exists()

    def test_list_projects_with_namespace_input(self):
        name_of_the_repo = "inkscape"
        number_of_projects = 10
        projects = self.service.list_projects(namespace=name_of_the_repo)
        assert len(projects) == number_of_projects

    def test_list_projects_with_namespace_input_and_language_input(self):
        name_of_the_repo = "inkscape"
        language = "C++"
        number_of_projects = 2
        projects = self.service.list_projects(
            namespace=name_of_the_repo,
            language=language,
        )
        assert len(projects) == number_of_projects

    def test_list_projects_get_forks(self):
        projects = self.service.list_projects(namespace="inkscape")
        assert projects
        assert isinstance(projects[0].get_forks(), list)

    def test_wrong_auth(self):
        with pytest.raises(GitlabAPIException):
            self.service.project_create("test")

    def test_wrong_auth_static_method(self):
        # Verify that the exception handler is applied to static methods
        with pytest.raises(GitlabAPIException):
            GitlabPullRequest.get(self.project, 1)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import Optional

from ogr.abstract import GitTag, Release
from ogr.exceptions import OperationNotSupported, PagureAPIException
from ogr.services import pagure as ogr_pagure


class PagureRelease(Release):
    _raw_release: GitTag
    project: "ogr_pagure.PagureProject"

    @property
    def title(self):
        return self.git_tag.name

    @property
    def body(self):
        return ""

    @property
    def git_tag(self) -> GitTag:
        return self._raw_release

    @property
    def tag_name(self) -> str:
        return self._raw_release.name

    @property
    def url(self) -> Optional[str]:
        return ""

    @property
    def created_at(self) -> datetime.datetime:
        return None

    @property
    def tarball_url(self) -> str:
        return ""

    def __str__(self) -> str:
        return "Pagure" + super().__str__()

    @staticmethod
    def get(
        project: "ogr_pagure.PagureProject",
        identifier: Optional[int] = None,
        name: Optional[str] = None,
        tag_name: Optional[str] = None,
    ) -> "Release":
        raise OperationNotSupported()

    @staticmethod
    def get_latest(project: "ogr_pagure.PagureProject") -> Optional["Release"]:
        raise OperationNotSupported("Pagure API does not provide timestamps")

    @staticmethod
    def get_list(project: "ogr_pagure.PagureProject") -> list["Release"]:
        # git tag for Pagure is shown as Release in Pagure UI
        git_tags = project.get_tags()
        return [PagureRelease(git_tag, project) for git_tag in git_tags]

    @staticmethod
    def create(
        project: "ogr_pagure.PagureProject",
        tag: str,
        name: str,
        message: str,
        ref: Optional[str] = None,
    ) -> "Release":
        payload = {
            "tagname": tag,
            "commit_hash": ref,
        }
        if message:
            payload["message"] = message

        response = project._call_project_api("git", "tags", data=payload, method="POST")
        if not response["tag_created"]:
            raise PagureAPIException("Release has not been created")

        return PagureRelease(GitTag(tag, ref), project)

    def edit_release(self, name: str, message: str) -> None:
        raise OperationNotSupported("edit_release not supported on Pagure")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import Any, ClassVar, Optional

from ogr.abstract import CommitFlag, CommitStatus
from ogr.services import pagure as ogr_pagure
from ogr.services.base import BaseCommitFlag


class PagureCommitFlag(BaseCommitFlag):
    _states: ClassVar[dict[str, CommitStatus]] = {
        "pending": CommitStatus.pending,
        "success": CommitStatus.success,
        "failure": CommitStatus.failure,
        "error": CommitStatus.error,
        "canceled": CommitStatus.canceled,
    }

    def __str__(self) -> str:
        return "Pagure" + super().__str__()

    def _from_raw_commit_flag(self):
        self.commit = self._raw_commit_flag["commit_hash"]
        self.comment = self._raw_commit_flag["comment"]
        self.state = self._state_from_str(self._raw_commit_flag["status"])
        self.context = self._raw_commit_flag["username"]
        self.url = self._raw_commit_flag["url"]

    @staticmethod
    def get(project: "ogr_pagure.PagureProject", commit: str) -> list["CommitFlag"]:
        response = project._call_project_api("c", commit, "flag")
        return [
            PagureCommitFlag(raw_commit_flag=flag, project=project)
            for flag in response["flags"]
        ]

    @staticmethod
    def set(
        project: "ogr_pagure.PagureProject",
        commit: str,
        state: CommitStatus,
        target_url: str,
        description: str,
        context: str,
        percent: Optional[int] = None,
        trim: bool = False,
        uid: Optional[str] = None,
    ) -> "CommitFlag":
        state = PagureCommitFlag._validate_state(state)

        if trim:
            description = description[:140]

        data: dict[str, Any] = {
            "username": context,
            "comment": description,
            "url": target_url,
            "status": state.name,
        }
        if percent:
            data["percent"] = percent
        if uid:
            data["uid"] = uid

        response = project._call_project_api(
            "c",
            commit,
            "flag",
            method="POST",
            data=data,
        )
        return PagureCommitFlag(
            project=project,
            raw_commit_flag=response["flag"],
            uid=response["uid"],
        )

    @property
    def created(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            int(self._raw_commit_flag["date_created"]),
        )

    @property
    def edited(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            int(self._raw_commit_flag["date_updated"]),
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.services.gitlab.comments import GitlabIssueComment, GitlabPRComment
from ogr.services.gitlab.issue import GitlabIssue
from ogr.services.gitlab.project import GitlabProject
from ogr.services.gitlab.pull_request import GitlabPullRequest
from ogr.services.gitlab.release import GitlabRelease
from ogr.services.gitlab.service import GitlabService
from ogr.services.gitlab.user import GitlabUser

__all__ = [
    GitlabIssue.__name__,
    GitlabPullRequest.__name__,
    GitlabIssueComment.__name__,
    GitlabPRComment.__name__,
    GitlabRelease.__name__,
    GitlabUser.__name__,
    GitlabProject.__name__,
    GitlabService.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import Optional

from gitlab.v4.objects import ProjectRelease as _GitlabRelease

from ogr.abstract import GitTag, Release
from ogr.exceptions import OperationNotSupported
from ogr.services import gitlab as ogr_gitlab


class GitlabRelease(Release):
    _raw_release: _GitlabRelease
    project: "ogr_gitlab.GitlabProject"

    @property
    def title(self):
        return self._raw_release.name

    @property
    def body(self):
        return self._raw_release.description

    @property
    def git_tag(self) -> GitTag:
        return self.project._git_tag_from_tag_name(self.tag_name)

    @property
    def tag_name(self) -> str:
        return self._raw_release.tag_name

    @property
    def url(self) -> Optional[str]:
        return f"{self.project.get_web_url()}/-/releases/{self.tag_name}"

    @property
    def created_at(self) -> datetime.datetime:
        return self._raw_release.created_at

    @property
    def tarball_url(self) -> str:
        return self._raw_release.assets["sources"][1]["url"]

    def __str__(self) -> str:
        return "Gitlab" + super().__str__()

    @staticmethod
    def get(
        project: "ogr_gitlab.GitlabProject",
        identifier: Optional[int] = None,
        name: Optional[str] = None,
        tag_name: Optional[str] = None,
    ) -> "Release":
        release = project.gitlab_repo.releases.get(tag_name)
        return GitlabRelease(release, project)

    @staticmethod
    def get_latest(project: "ogr_gitlab.GitlabProject") -> Optional["Release"]:
        releases = project.gitlab_repo.releases.list()
        # list of releases sorted by released_at
        return GitlabRelease(releases[0], project) if releases else None

    @staticmethod
    def get_list(project: "ogr_gitlab.GitlabProject") -> list["Release"]:
        if not hasattr(project.gitlab_repo, "releases"):
            raise OperationNotSupported(
                "This version of python-gitlab does not support release, please upgrade.",
            )
        releases = project.gitlab_repo.releases.list(all=True)
        return [GitlabRelease(release, project) for release in releases]

    @staticmethod
    def create(
        project: "ogr_gitlab.GitlabProject",
        tag: str,
        name: str,
        message: str,
        ref: Optional[str] = None,
    ) -> "Release":
        release = project.gitlab_repo.releases.create(
            {"name": name, "tag_name": tag, "description": message, "ref": ref},
        )
        return GitlabRelease(release, project)

    def edit_release(self, name: str, message: str) -> None:
        raise OperationNotSupported("edit_release not supported on GitLab")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import logging
from typing import ClassVar

import gitlab

from ogr.abstract import CommitFlag, CommitStatus
from ogr.exceptions import GitlabAPIException, OperationNotSupported
from ogr.services import gitlab as ogr_gitlab
from ogr.services.base import BaseCommitFlag

logger = logging.getLogger(__name__)


class GitlabCommitFlag(BaseCommitFlag):
    _states: ClassVar[dict[str, CommitStatus]] = {
        "pending": CommitStatus.pending,
        "success": CommitStatus.success,
        "failed": CommitStatus.failure,
        "canceled": CommitStatus.canceled,
        "running": CommitStatus.running,
    }

    @staticmethod
    def _state_from_enum(state: CommitStatus) -> str:
        return "failed" if state == CommitStatus.failure else state.name

    def __str__(self) -> str:
        return (
            f"GitlabCommitFlag("
            f"commit='{self.commit}', "
            f"state='{self.state.name}', "
            f"context='{self.context}', "
            f"uid='{self.uid}', "
            f"comment='{self.comment}', "
            f"url='{self.url}', "
            f"created='{self.created}')"
        )

    def _from_raw_commit_flag(self):
        self.commit = self._raw_commit_flag.sha
        self.state = self._state_from_str(self._raw_commit_flag.status)
        self.context = self._raw_commit_flag.name
        self.comment = self._raw_commit_flag.description
        self.uid = self._raw_commit_flag.id
        self.url = self._raw_commit_flag.target_url

    @staticmethod
    def get(project: "ogr_gitlab.GitlabProject", commit: str) -> list["CommitFlag"]:
        try:
            commit_object = project.gitlab_repo.commits.get(commit)
        except gitlab.exceptions.GitlabGetError as ex:
            logger.error(f"Commit {commit} was not found.")
            raise GitlabAPIException(f"Commit {commit} was not found.") from ex

        raw_statuses = commit_object.statuses.list(all=True)
        return [
            GitlabCommitFlag(raw_commit_flag=raw_status, project=project)
            for raw_status in raw_statuses
        ]

    @staticmethod
    def set(
        project: "ogr_gitlab.GitlabProject",
        commit: str,
        state: CommitStatus,
        target_url: str,
        description: str,
        context: str,
        trim: bool = False,
    ) -> "CommitFlag":
        state = GitlabCommitFlag._validate_state(state)

        if trim:
            description = description[:140]

        try:
            commit_object = project.gitlab_repo.commits.get(commit)
        except gitlab.exceptions.GitlabGetError as ex:
            logger.error(f"Commit {commit} was not found.")
            raise GitlabAPIException(f"Commit {commit} was not found.") from ex

        data_dict = {
            "state": GitlabCommitFlag._state_from_enum(state),
            "target_url": target_url,
            "context": context,
            "description": description,
        }
        raw_status = commit_object.statuses.create(data_dict)
        return GitlabCommitFlag(raw_commit_flag=raw_status, project=project)

    @property
    def created(self) -> datetime.datetime:
        return datetime.datetime.strptime(
            self._raw_commit_flag.created_at,
            "%Y-%m-%dT%H:%M:%S.%fZ",
        )

    @property
    def edited(self) -> datetime.datetime:
        raise OperationNotSupported(
            "GitLab doesn't support edited on commit flags, for more info "
            "see https://github.com/packit/ogr/issues/413#issuecomment-729623702",
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from collections import namedtuple
from typing import Optional

from ogr.services import github as ogr_github
from ogr.services.base import BaseGitUser
from ogr.services.github.project import GithubProject


class GithubUser(BaseGitUser):
    service: "ogr_github.GithubService"

    def __init__(self, service: "ogr_github.GithubService") -> None:
        super().__init__(service=service)

    def __str__(self) -> str:
        return f'GithubUser(username="{self.get_username()}")'

    @property
    def _github_user(self):
        return self.service.github.get_user()

    def get_username(self) -> str:
        return self.service.github.get_user().login

    def get_email(self) -> Optional[str]:
        user_email_property = self.service.github.get_user().email
        if user_email_property:
            return user_email_property

        user_emails = self.service.github.get_user().get_emails()

        if not user_emails:
            return None

        # To work around the braking change introduced by pygithub==1.55
        # https://pygithub.readthedocs.io/en/latest/changes.html#version-1-55-april-26-2021
        if isinstance(user_emails[0], dict):
            EmailData = namedtuple("EmailData", user_emails[0].keys())  # type: ignore
        for email in user_emails:
            if "EmailData" in locals():
                email = EmailData(**email)  # type: ignore
            if email.primary:
                return email.email

        # Return the first email we received
        return user_emails[0]["email"]

    def get_projects(self) -> list["ogr_github.GithubProject"]:
        raw_repos = self._github_user.get_repos(affiliation="owner")
        return [
            GithubProject(
                repo=repo.name,
                namespace=repo.owner.login,
                github_repo=repo,
                service=self.service,
            )
            for repo in raw_repos
        ]

    def get_forks(self) -> list["ogr_github.GithubProject"]:
        return [project for project in self.get_projects() if project.github_repo.fork]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
from datetime import datetime

import pytest
from requre.online_replacing import record_requests_for_all_methods
from requre.utils import get_datafile_filename

from ogr import PagureService
from ogr.abstract import CommitStatus, IssueStatus
from ogr.exceptions import PagureAPIException
from tests.integration.pagure.base import PagureTests


@record_requests_for_all_methods()
class PagureProjectTokenCommands(PagureTests):
    def setUp(self):
        super().setUp()
        self.token = os.environ.get("PAGURE_OGR_TEST_TOKEN", "")

        if not get_datafile_filename(obj=self) and (not self.token):
            raise OSError("please set PAGURE_OGR_TEST_TOKEN env variables")

        self._service = None
        self._user = None
        self._ogr_project = None
        self._ogr_fork = None

    @property
    def service(self):
        if not self._service:
            self._service = PagureService(
                token=self.token,
                instance_url="https://pagure.io",
            )
        return self._service

    def test_issue_permissions(self):
        owners = self.ogr_project.who_can_close_issue()
        assert "lachmanfrantisek" in owners

        issue = self.ogr_project.get_issue(2)
        assert issue.can_close("lachmanfrantisek")

    def test_issue_comments(self):
        issue_comments = self.ogr_project.get_issue(3)._get_all_comments()
        assert issue_comments
        assert len(issue_comments) == 4
        assert issue_comments[0].body.startswith("test")
        assert issue_comments[1].body.startswith("tests")

    def test_issue_info(self):
        issue_info = self.ogr_project.get_issue(issue_id=2)
        assert issue_info
        assert issue_info.title.startswith("Test 1")
        assert issue_info.status == IssueStatus.closed

    def test_issue_comments_reversed(self):
        issue_comments = self.ogr_project.get_issue(3).get_comments(reverse=True)
        assert len(issue_comments) == 4
        assert issue_comments[0].body.startswith("regex")

    def test_issue_comments_regex(self):
        issue_comments = self.ogr_project.get_issue(3).get_comments(
            filter_regex="regex",
        )
        assert len(issue_comments) == 2
        assert issue_comments[0].body.startswith("let's")

    def test_issue_comments_regex_reversed(self):
        issue_comments = self.ogr_project.get_issue(3).get_comments(
            filter_regex="regex",
            reverse=True,
        )
        assert len(issue_comments) == 2
        assert issue_comments[0].body.startswith("regex")

    def test_issue_update_title(self):
        issue = self.ogr_project.get_issue(3)
        old_title, old_description = issue.title, issue.description

        issue.title = "testing title"
        assert (issue.title, issue.description) == ("testing title", old_description)

        issue.title = old_title
        assert (issue.title, issue.description) == (old_title, old_description)

    def test_issue_update_description(self):
        issue = self.ogr_project.get_issue(3)
        old_title, old_description = issue.title, issue.description

        issue.description = "testing description"
        assert (issue.title, issue.description) == (old_title, "testing description")

        issue.description = old_description
        assert (issue.title, issue.description) == (old_title, old_description)

    def test_update_pr_info(self):
        pr_info = self.ogr_project.get_pr(pr_id=4)
        orig_title = pr_info.title
        orig_description = pr_info.description

        self.ogr_project.get_pr(4).update_info(
            title="changed",
            description="changed description",
        )
        pr_info = self.ogr_project.get_pr(pr_id=4)
        assert pr_info.title == "changed"
        assert pr_info.description == "changed description"

        self.ogr_project.get_pr(4).update_info(
            title=orig_title,
            description=orig_description,
        )
        pr_info = self.ogr_project.get_pr(pr_id=4)
        assert pr_info.title == orig_title
        assert pr_info.description == orig_description

    def test_pr_setters(self):
        pr = self.ogr_project.get_pr(pr_id=6)

        old_title = pr.title
        pr.title = "test title"
        assert pr.title != old_title
        assert pr.title == "test title"

        pr.title = old_title
        assert pr.title == old_title

        old_description = pr.description
        pr.description = "test description"
        assert pr.description != old_description
        assert pr.description == "test description"

        pr.description = old_description
        assert pr.description == old_description

    def test_pr_comments_author_regex(self):
        comments = self.ogr_project.get_pr(4).get_comments(
            filter_regex="^regex",
            author="mfocko",
        )
        assert len(comments) == 1
        assert comments[0].body.endswith("test")

    def test_pr_comments_author(self):
        comments = self.ogr_project.get_pr(4).get_comments(author="lachmanfrantisek")
        assert len(comments) == 0

    def test_issue_comments_author_regex(self):
        comments = self.ogr_project.get_issue(3).get_comments(
            filter_regex="^test[s]?$",
            author="mfocko",
        )
        assert len(comments) == 2
        assert comments[0].body == "test"
        assert comments[1].body == "tests"

    def test_issue_comments_author(self):
        comments = self.ogr_project.get_issue(3).get_comments(author="lachmanfrantisek")
        assert len(comments) == 0

    def test_pr_status(self):
        pr = self.ogr_project.get_pr(pr_id=4)
        self.ogr_project.set_commit_status(
            commit=pr.head_commit,
            state=CommitStatus.success,
            target_url="https://pagure.io/ogr-tests/pull-request/4",
            description="not failed test",
            context="test",
        )

        statuses = pr.get_statuses()
        assert statuses
        assert len(statuses) >= 0
        assert statuses[-1].state == CommitStatus.success
        # What timezone?
        assert statuses[-1].created >= datetime(
            year=2020,
            month=8,
            day=31,
            hour=1,
            minute=0,
            second=0,
        )
        assert statuses[-1].edited >= datetime(
            year=2020,
            month=8,
            day=31,
            hour=1,
            minute=0,
            second=0,
        )

    def test_is_private(self):
        self.service.instance_url = "https://src.fedoraproject.org"
        assert not self.ogr_project.is_private()

    def test_token_is_none_then_set(self):
        token = self.service._token
        self.service.change_token("")
        try:
            with pytest.raises(PagureAPIException) as exc:
                self.service.user.get_username()
            assert "Invalid or expired token" in str(exc)
        finally:
            self.service.change_token(token)

        self.service.user.get_username()
        self.service.user.get_username()  # 2nd identical call

    def test_create_release(self):
        self.ogr_project.create_release(
            tag="v0",
            name="v0",
            message="# v0\n\n• added README",
            ref="2988640e03ddee8385a2acb827a36c8e50b1be1a",
        )
        assert "v0" in (release.tag_name for release in self.ogr_project.get_releases())



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from requre.online_replacing import record_requests_for_all_methods

from ogr.abstract import IssueStatus
from ogr.exceptions import (
    GithubAPIException,
    IssueTrackerDisabled,
    OperationNotSupported,
)
from tests.integration.github.base import GithubTests


@record_requests_for_all_methods()
class Issues(GithubTests):
    title = "This is a title"
    description = "This is a description"

    def test_issue_list(self):
        with pytest.raises(IssueTrackerDisabled):
            issue_list = self.ogr_fork.get_issue_list()

        issue_list_all = self.ogr_project.get_issue_list(status=IssueStatus.all)
        assert issue_list_all
        assert len(issue_list_all) >= 45

        issue_list_closed = self.ogr_project.get_issue_list(status=IssueStatus.closed)
        assert issue_list_closed
        assert len(issue_list_closed) >= 35

        issue_list = self.ogr_project.get_issue_list()
        assert issue_list
        assert len(issue_list) >= 3

    def test_create_issue(self):
        labels = ["label1", "label2"]
        issue = self.hello_world_project.create_issue(
            title=self.title,
            body=self.description,
            labels=labels,
        )
        assert issue.title == self.title
        assert issue.description == self.description
        for issue_label, label in zip(issue.labels, labels):
            assert issue_label.name == label

    def test_create_private_issue(self):
        with pytest.raises(OperationNotSupported):
            self.hello_world_project.create_issue(
                title=self.title,
                body=self.description,
                private=True,
            )

    def test_create_issue_with_assignee(self):
        labels = ["label1", "label2"]
        assignee = ["lachmanfrantisek"]
        issue = self.hello_world_project.create_issue(
            title=self.title,
            body=self.description,
            labels=labels,
            assignees=assignee,
        )
        assert issue.title == self.title
        assert issue.description == self.description
        assert issue.assignees[0].login == assignee[0]

    def test_issue_without_label(self):
        project = self.service.get_project(namespace="shreyaspapi", repo="test")
        title = "This is an issue"
        description = "Example of Issue description"
        issue = project.create_issue(title=title, body=description)
        assert issue.title == title
        assert issue.description == description

    def test_issue_list_author(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            author="mfocko",
        )
        assert issue_list
        assert len(issue_list) >= 12

    def test_issue_list_nonexisting_author(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            author="xyzidontexist",
        )
        assert len(issue_list) == 0

    def test_issue_list_assignee(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            assignee="mfocko",
        )
        assert issue_list
        assert len(issue_list) >= 10

    def test_issue_list_labels(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            labels=["Pagure"],
        )
        assert issue_list
        assert len(issue_list) >= 42

    def test_issue_info(self):
        issue_info = self.ogr_project.get_issue(4)
        assert issue_info
        assert issue_info.title.startswith("Better name")
        assert issue_info.status == IssueStatus.closed

    def test_issue_labels(self):
        """
        Remove the labels from this issue before regenerating the response files:
        https://github.com/packit/ogr/issues/4
        """
        issue = self.ogr_project.get_issue(4)
        labels = issue.labels

        assert not labels
        issue.add_label("test_lb1", "test_lb2")
        labels = self.ogr_project.get_issue(4).labels
        assert len(labels) == 2
        assert labels[0].name == "test_lb1"
        assert labels[1].name == "test_lb2"

    def test_issue_assignees(self):
        """
        Remove the assignees from this issue before regenerating the response files:
        https://github.com/packit/ogr/issues/4
        """
        project = self.service.get_project(repo="ogr", namespace="KPostOffice")
        issue = project.get_issue(4)
        print(self.service.user.get_username())
        assignees = issue.assignees

        assert not assignees
        issue.add_assignee("KPostOffice")
        assignees = project.get_issue(4).assignees
        assert len(assignees) == 1
        assert assignees[0].login == "KPostOffice"

    def test_issue_no_such_assignee(self):
        project = self.service.get_project(repo="ogr", namespace="KPostOffice")
        issue = project.get_issue(4)
        with pytest.raises(GithubAPIException):
            issue.add_assignee("slkdfjglkslfslkabcd")

    def test_list_contains_only_issues(self):
        issue_list_all = self.ogr_project.get_issue_list(status=IssueStatus.all)
        issue_ids = [issue.id for issue in issue_list_all]

        pr_ids = [219, 207, 201, 217, 208, 210]
        for id in pr_ids:
            assert id not in issue_ids

    def test_issue_not_exists(self):
        with pytest.raises(GithubAPIException):
            self.ogr_project.get_issue(10**20)

    def test_functions_fail_for_pr(self):
        with pytest.raises(GithubAPIException):
            self.ogr_project.get_issue(1)
        with pytest.raises(GithubAPIException):
            self.ogr_project.get_issue(1).comment(body="should fail")
        with pytest.raises(GithubAPIException):
            self.ogr_project.get_issue(1).close()
        with pytest.raises(GithubAPIException):
            _ = self.ogr_project.get_issue(1).labels
        with pytest.raises(GithubAPIException):
            self.ogr_project.get_issue(1).add_label("should fail")

    def test_issue_updates(self):
        issue = self.hello_world_project.get_issue(issue_id=61)
        old_comments = issue.get_comments()
        issue.comment("test comment")
        new_comments = issue.get_comments()
        assert len(new_comments) > len(old_comments)

    def test_setters(self):
        issue = self.hello_world_project.get_issue(issue_id=61)

        old_title = issue.title
        issue.title = "test title"
        assert issue.title != old_title
        assert issue.title == "test title"

        issue.title = old_title
        assert issue.title == old_title

        old_description = issue.description
        issue.description = "test description"
        assert issue.description != old_description
        assert issue.description == "test description"

        issue.description = old_description
        assert issue.description == old_description

    def test_get_comment(self):
        comment = self.hello_world_project.get_issue(297).get_comment(926035728)
        assert comment.body == "this is a comment"

    def test_create_with_disabled_issues(self):
        with pytest.raises(IssueTrackerDisabled):
            self.hello_world_project.get_fork().create_issue(
                "Testing issue",
                "shouldn't be created",
            )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import unittest
from pathlib import Path

from requre.utils import get_datafile_filename

from ogr.services.gitlab import GitlabService


class GitlabTests(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.token = os.environ.get("GITLAB_TOKEN")

        if not Path(get_datafile_filename(obj=self)).exists() and not self.token:
            raise OSError(
                "You are in Requre write mode, please set GITLAB_TOKEN env variables",
            )

        if not self.token:
            self.token = "some_token"

        self._service = None
        self._project = None

    @property
    def service(self):
        if not self._service:
            self._service = GitlabService(
                token=self.token,
                instance_url="https://gitlab.com",
                ssl_verify=True,
            )
        return self._service

    @property
    def project(self):
        if not self._project:
            self._project = self.service.get_project(
                repo="ogr-tests",
                namespace="packit-service",
            )
        return self._project



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from ogr.parsing import RepoUrl, parse_git_repo


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "https://host.name/namespace/repo",
            RepoUrl(
                repo="repo",
                namespace="namespace",
                scheme="https",
                hostname="host.name",
                username="namespace",
            ),
        ),
        (
            "https://host.name/namespace/repo.git",
            RepoUrl(
                repo="repo",
                namespace="namespace",
                scheme="https",
                hostname="host.name",
                username="namespace",
            ),
        ),
        (
            "http://host.name/namespace/repo",
            RepoUrl(
                repo="repo",
                namespace="namespace",
                scheme="http",
                hostname="host.name",
                username="namespace",
            ),
        ),
        (
            "git://host.name/namespace/repo",
            RepoUrl(
                repo="repo",
                namespace="namespace",
                scheme="git",
                hostname="host.name",
                username="namespace",
            ),
        ),
        (
            "git+https://host.name/namespace/repo",
            RepoUrl(
                repo="repo",
                namespace="namespace",
                scheme="git+https",
                hostname="host.name",
                username="namespace",
            ),
        ),
        (
            "git@host.name:namespace/repo",
            RepoUrl(
                repo="repo",
                namespace="namespace",
                scheme="https",
                hostname="host.name",
                username="namespace",
            ),
        ),
        ("host.name/repo", RepoUrl(repo="repo", scheme="https", hostname="host.name")),
        (
            "host.name/fork/user/namespace/repo",
            RepoUrl(
                repo="repo",
                username="user",
                namespace="namespace",
                scheme="https",
                hostname="host.name",
                is_fork=True,
            ),
        ),
        (
            "https://host.name/namespace/repo/",
            RepoUrl(
                repo="repo",
                username="namespace",
                namespace="namespace",
                scheme="https",
                hostname="host.name",
            ),
        ),
        (
            "https://host.name/multi/part/namespace/repo/",
            RepoUrl(
                repo="repo",
                username="multi",
                namespace="multi/part/namespace",
                scheme="https",
                hostname="host.name",
            ),
        ),
        (
            "https://pagure.io/fork/user/some_repo",
            RepoUrl(
                repo="some_repo",
                username="user",
                namespace="",
                is_fork=True,
                hostname="pagure.io",
                scheme="https",
            ),
        ),
        ("https://fail@more@at@domain.com", None),
        (
            "git@gitlab.com:packit-service/src/libvirt.git",
            RepoUrl(
                repo="libvirt",
                namespace="packit-service/src",
                username="packit-service",
                hostname="gitlab.com",
                scheme="https",
            ),
        ),
        ("git@git.mfocko.xyz:2222:mfocko/dotfiles.git", None),
        (
            "https://pagure.io/fork/mfocko/fedora-infra/ansible.git",
            RepoUrl(
                repo="ansible",
                namespace="fedora-infra",
                username="mfocko",
                is_fork=True,
                hostname="pagure.io",
                scheme="https",
            ),
        ),
        (
            "ssh://git@pagure.io/forks/mfocko/fedora-infra/ansible.git",
            RepoUrl(
                repo="ansible",
                namespace="fedora-infra",
                username="mfocko",
                is_fork=True,
                hostname="pagure.io",
                scheme="https",
            ),
        ),
        (
            "ssh://git@pagure.io:forks/mfocko/fedora-infra/ansible.git",
            RepoUrl(
                repo="ansible",
                namespace="fedora-infra",
                username="mfocko",
                is_fork=True,
                hostname="pagure.io",
                scheme="https",
            ),
        ),
        (
            "https://xfocko:myLamePassword@git.mfocko.xyz/mfocko/dotfiles.git",
            RepoUrl(
                repo="dotfiles",
                namespace="mfocko",
                hostname="git.mfocko.xyz",
                scheme="https",
                username="mfocko",
            ),
        ),
        (
            "ssh://git@pagure.io/playground-mfocko.git",
            RepoUrl(
                repo="playground-mfocko",
                namespace=None,
                hostname="pagure.io",
                scheme="https",
            ),
        ),
        (
            "https://pagure.io/playground-mfocko.git",
            RepoUrl(
                repo="playground-mfocko",
                namespace=None,
                hostname="pagure.io",
                scheme="https",
            ),
        ),
        (
            "git://github.com/packit/dotfiles.git",
            RepoUrl(
                repo="dotfiles",
                namespace="packit",
                username="packit",
                hostname="github.com",
                scheme="git",
            ),
        ),
        (
            "https://git.launchpad.net/dkimpy",
            RepoUrl(
                repo="dkimpy",
                hostname="git.launchpad.net",
                scheme="https",
            ),
        ),
    ],
)
def test_parse_git_repo(url, expected):
    repo_url = parse_git_repo(potential_url=url)
    assert repo_url == expected



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from datetime import datetime

import pytest
from requre.online_replacing import record_requests_for_all_methods

from ogr.exceptions import OperationNotSupported
from ogr.services.github import GithubProject
from ogr.services.github.check_run import (
    GithubCheckRunResult,
    GithubCheckRunStatus,
    create_github_check_run_output,
)
from tests.integration.github.base_app import GithubAppTests


@record_requests_for_all_methods()
class CheckRun(GithubAppTests):
    @property
    def project(self) -> GithubProject:
        return self.hello_world_project

    def test_non_existing_check_runs_returns_none(self):
        check_run = self.project.get_check_run(
            commit_sha="f502aae6920d82948f2dba0b70c9260fb1e34822",
        )

        assert check_run is None

    def test_get_list(self):
        check_runs = self.project.get_check_runs(
            "7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
        )

        assert check_runs
        assert len(check_runs) >= 3

    def test_get_list_no_runs(self):
        check_runs = self.project.get_check_runs(
            "f502aae6920d82948f2dba0b70c9260fb1e34822",
        )
        assert check_runs == []

    def test_create_to_queue_and_succeed(self):
        check_run = self.project.create_check_run(
            name="check run to be queued",
            commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
            url="https://localhost",
            external_id="ogr-test",
        )

        assert check_run
        assert check_run.name == "check run to be queued"

        check_run.change_status(conclusion=GithubCheckRunResult.success)

        assert check_run.status == GithubCheckRunStatus.completed
        assert check_run.conclusion == GithubCheckRunResult.success

    def test_create_neutral_completed(self):
        check_run = self.project.create_check_run(
            name="neutral completed",
            commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
            url="https://localhost",
            external_id="ogr-test",
            conclusion=GithubCheckRunResult.neutral,
            output=create_github_check_run_output(
                "Compile & run tests",
                "Compilation failed",
                (
                    "# Compile\n\n**FAILED**\n```\nunused parameters\n```\n\n"
                    "# Tests\n\ndependency failed"
                ),
            ),
        )

        assert check_run
        assert check_run.conclusion == GithubCheckRunResult.neutral

    def test_create_timed_out(self):
        check_run = self.project.create_check_run(
            name="timed out",
            commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
            url="https://localhost",
            external_id="ogr-test",
            conclusion=GithubCheckRunResult.timed_out,
            output=create_github_check_run_output(
                "Compile & run tests",
                "Tests timed out",
            ),
        )

        assert check_run
        assert check_run.conclusion == GithubCheckRunResult.timed_out

    def test_create_with_completed_without_conclusion(self):
        with pytest.raises(OperationNotSupported):
            self.project.create_check_run(
                name="should fail",
                commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
                status=GithubCheckRunStatus.completed,
            )

    def test_create_with_completed_at_without_conclusion(self):
        with pytest.raises(OperationNotSupported):
            self.project.create_check_run(
                name="should fail",
                commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
                completed_at=datetime(
                    year=2021,
                    month=5,
                    day=27,
                    hour=11,
                    minute=11,
                    second=11,
                ),
            )

    def test_create_completed_without_conclusion(self):
        with pytest.raises(OperationNotSupported):
            self.project.create_check_run(
                name="should fail",
                commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
                status=GithubCheckRunStatus.completed,
                completed_at=datetime(
                    year=2021,
                    month=5,
                    day=27,
                    hour=11,
                    minute=11,
                    second=11,
                ),
            )

    # these tests need to have at least one check run on given commit
    def test_get_latest_check_run(self):
        check_run = self.project.get_check_run(
            commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
        )
        assert check_run

    def test_change_name(self):
        check_run = self.project.get_check_run(
            commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
        )
        assert check_run, "No check run exists"

        check_run.name = "New check run name"
        assert check_run.name == "New check run name"

    def test_change_url(self):
        check_run = self.project.get_check_run(
            commit_sha="7cf6d0cbeca285ecbeb19a0067cb243783b3c768",
        )

        assert check_run, "No check run exists"

        check_run.url = "https://packit.dev"
        assert check_run.url == "https://packit.dev"

        check_run.url = "https://dashboard.packit.dev"
        assert check_run.url == "https://dashboard.packit.dev"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging

from ogr.services import pagure as ogr_pagure

logger = logging.getLogger(__name__)


class PagureGroup:
    service: "ogr_pagure.PagureService"

    def __init__(self, name: str, raw_group: dict) -> None:
        self.name = name
        # see https://pagure.io/api/0/#groups-tab
        self._raw_group = raw_group

    def __str__(self) -> str:
        return f'PagureGroup(name="{self.name}")'

    @property
    def members(self) -> list[str]:
        return self._raw_group["members"]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import Any, Optional

from ogr.abstract import Comment, IssueComment, PRComment
from ogr.exceptions import OperationNotSupported


class PagureComment(Comment):
    def _from_raw_comment(self, raw_comment: dict[str, Any]) -> None:
        self._body = raw_comment["comment"]
        self._id = raw_comment["id"]
        self._author = raw_comment["user"]["name"]
        self._created = self.__datetime_from_timestamp(raw_comment["date_created"])
        self._edited = self.__datetime_from_timestamp(raw_comment["edited_on"])

    @staticmethod
    def __datetime_from_timestamp(
        timestamp: Optional[str],
    ) -> Optional[datetime.datetime]:
        return datetime.datetime.fromtimestamp(int(timestamp)) if timestamp else None

    @property
    def body(self) -> str:
        return self._body

    @body.setter
    def body(self, new_body: str) -> None:
        raise OperationNotSupported()


class PagureIssueComment(PagureComment, IssueComment):
    def __str__(self) -> str:
        return "Pagure" + super().__str__()


class PagurePRComment(PagureComment, PRComment):
    def __str__(self) -> str:
        return "Pagure" + super().__str__()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import Optional, Union

import github
from github import UnknownObjectException
from github.Issue import Issue as _GithubIssue

from ogr.abstract import Issue, IssueComment, IssueLabel, IssueStatus
from ogr.exceptions import (
    GithubAPIException,
    IssueTrackerDisabled,
    OperationNotSupported,
)
from ogr.services import github as ogr_github
from ogr.services.base import BaseIssue
from ogr.services.github.comments import GithubIssueComment
from ogr.services.github.label import GithubIssueLabel


class GithubIssue(BaseIssue):
    raw_issue: _GithubIssue

    def __init__(
        self,
        raw_issue: _GithubIssue,
        project: "ogr_github.GithubProject",
    ) -> None:
        if raw_issue.pull_request:
            raise GithubAPIException(
                f"Requested issue #{raw_issue.number} is a pull request",
            )

        super().__init__(raw_issue=raw_issue, project=project)

    @property
    def title(self) -> str:
        return self._raw_issue.title

    @title.setter
    def title(self, new_title: str) -> None:
        self._raw_issue.edit(title=new_title)

    @property
    def id(self) -> int:
        return self._raw_issue.number

    @property
    def status(self) -> IssueStatus:
        return IssueStatus[self._raw_issue.state]

    @property
    def url(self) -> str:
        return self._raw_issue.html_url

    @property
    def assignees(self) -> list:
        return self._raw_issue.assignees

    @property
    def description(self) -> str:
        return self._raw_issue.body

    @description.setter
    def description(self, new_description: str) -> None:
        self._raw_issue.edit(body=new_description)

    @property
    def author(self) -> str:
        return self._raw_issue.user.login

    @property
    def created(self) -> datetime.datetime:
        return self._raw_issue.created_at

    @property
    def labels(self) -> list[IssueLabel]:
        return [
            GithubIssueLabel(raw_label, self)
            for raw_label in self._raw_issue.get_labels()
        ]

    def __str__(self) -> str:
        return "Github" + super().__str__()

    @staticmethod
    def create(
        project: "ogr_github.GithubProject",
        title: str,
        body: str,
        private: Optional[bool] = None,
        labels: Optional[list[str]] = None,
        assignees: Optional[list] = None,
    ) -> "Issue":
        if private:
            raise OperationNotSupported("Private issues are not supported by Github")
        if not project.has_issues:
            raise IssueTrackerDisabled()

        github_issue = project.github_repo.create_issue(
            title=title,
            body=body,
            labels=labels or [],
            assignees=assignees or [],
        )
        return GithubIssue(github_issue, project)

    @staticmethod
    def get(project: "ogr_github.GithubProject", issue_id: int) -> "Issue":
        if not project.has_issues:
            raise IssueTrackerDisabled()

        try:
            issue = project.github_repo.get_issue(number=issue_id)
        except github.UnknownObjectException as ex:
            raise GithubAPIException(f"No issue with id {issue_id} found") from ex
        return GithubIssue(issue, project)

    @staticmethod
    def get_list(
        project: "ogr_github.GithubProject",
        status: IssueStatus = IssueStatus.open,
        author: Optional[str] = None,
        assignee: Optional[str] = None,
        labels: Optional[list[str]] = None,
    ) -> list["Issue"]:
        if not project.has_issues:
            raise IssueTrackerDisabled()

        parameters: dict[str, Union[str, list[str]]] = {
            "state": status.name,
            "sort": "updated",
            "direction": "desc",
        }
        if author:
            parameters["creator"] = author
        if assignee:
            parameters["assignee"] = assignee
        if labels:
            parameters["labels"] = [
                project.github_repo.get_label(label) for label in labels
            ]

        issues = project.github_repo.get_issues(**parameters)
        try:
            return [
                GithubIssue(issue, project)
                for issue in issues
                if not issue.pull_request
            ]
        except UnknownObjectException:
            return []

    def _get_all_comments(self) -> list[IssueComment]:
        return [
            GithubIssueComment(parent=self, raw_comment=raw_comment)
            for raw_comment in self._raw_issue.get_comments()
        ]

    def comment(self, body: str) -> IssueComment:
        comment = self._raw_issue.create_comment(body)
        return GithubIssueComment(parent=self, raw_comment=comment)

    def close(self) -> "Issue":
        self._raw_issue.edit(state="closed")
        return self

    def add_label(self, *labels: str) -> None:
        for label in labels:
            self._raw_issue.add_to_labels(label)

    def add_assignee(self, *assignees: str) -> None:
        try:
            self._raw_issue.edit(assignees=list(assignees))
        except github.GithubException as ex:
            raise GithubAPIException("Failed to assign issue, unknown user") from ex

    def get_comment(self, comment_id: int) -> IssueComment:
        return GithubIssueComment(self._raw_issue.get_comment(comment_id))



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Any, Optional

import github
import gitlab


class OgrException(Exception):
    """Something went wrong during our execution."""


class APIException(OgrException):
    """Generic API exception."""

    @property
    def response_code(self):
        raise NotImplementedError()


class PagureAPIException(APIException):
    """Exception related to Pagure API."""

    def __init__(
        self,
        *args: Any,
        pagure_error: Optional[str] = None,
        pagure_response: Optional[dict[str, Any]] = None,
        response_code: Optional[int] = None,
    ) -> None:
        super().__init__(*args)
        self._pagure_error = pagure_error
        self.pagure_response = pagure_response
        self._response_code = response_code

    @property
    def pagure_error(self):
        return self._pagure_error or self.__cause__

    @property
    def response_code(self):
        return self._response_code


class GithubAPIException(APIException):
    """Exception related to Github API."""

    @property
    def response_code(self):
        if self.__cause__ is None or not isinstance(
            self.__cause__,
            github.GithubException,
        ):
            return None
        return self.__cause__.status


class GitlabAPIException(APIException):
    """Exception related to Gitlab API."""

    @property
    def response_code(self):
        if self.__cause__ is None or not isinstance(self.__cause__, gitlab.GitlabError):
            return None
        return self.__cause__.response_code


class OperationNotSupported(OgrException):
    """Raise when the operation is not supported by the backend."""


class IssueTrackerDisabled(OperationNotSupported):
    """Issue tracker on the project is not enabled."""


class OgrNetworkError(OgrException):
    """Exception raised when an unexpected network error occurs."""


class GitForgeInternalError(OgrNetworkError):
    """Exception raised when git forge returns internal failure."""


class GithubAppNotInstalledError(OgrException):
    """Exception raised when GitHub App is not installed."""



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from enum import Enum
from typing import Any, Optional, Union

from github.CheckRun import CheckRun
from github.CheckRunOutput import CheckRunOutput
from github.GithubApp import GithubApp
from github.GithubObject import NotSet

from ogr.abstract import OgrAbstractClass
from ogr.exceptions import OperationNotSupported
from ogr.services import github as ogr_github

GithubCheckRunOutput = dict[str, Union[str, list[dict[str, Union[str, int]]]]]


class GithubCheckRunStatus(Enum):
    """
    Represents statuses GitHub check run can have.
    """

    queued = "queued"
    in_progress = "in_progress"
    completed = "completed"


class GithubCheckRunResult(Enum):
    """
    Represents conclusion/result of the GitHub check run.
    """

    action_required = "action_required"
    cancelled = "cancelled"
    failure = "failure"
    neutral = "neutral"
    success = "success"
    skipped = "skipped"
    stale = "stale"
    timed_out = "timed_out"


def value_or_NotSet(value: Optional[Any]) -> Any:
    """
    Wrapper for PyGithub, allows us to transform `None` into PyGithub's `NotSet`.

    Args:
        value: Value that can be None.

    Returns:
        If value is not None, value is returned; NotSet otherwise.
    """
    return value if value is not None else NotSet


def create_github_check_run_output(
    title: str,
    summary: str,
    text: Optional[str] = None,
    annotations: Optional[list[dict[str, Union[str, int]]]] = None,
) -> GithubCheckRunOutput:
    """
    Helper function for constructing valid GitHub output for check run.

    Args:
        title: Title of the output.
        summary: Summary of the output.
        text: Optional text for the output. Can be markdown formatted.

            Defaults to `None`.
        annotations: Optional annotations that are tied to source code.

    Returns:
        Dictionary that represents valid output for check run.
    """
    output: GithubCheckRunOutput = {
        "title": title,
        "summary": summary,
    }

    if text is not None:
        output["text"] = text

    if annotations is not None:
        output["annotations"] = annotations

    return output


class GithubCheckRun(OgrAbstractClass):
    def __init__(
        self,
        project: "ogr_github.GithubProject",
        raw_check_run: CheckRun,
    ) -> None:
        self.raw_check_run = raw_check_run
        self.project = project

    def __str__(self) -> str:
        return (
            f"GithubCheckRun(project={self.project}, name='{self.name}', "
            f"commit_sha='{self.commit_sha}', "
            f"url='{self.url}', "
            f"external_id='{self.external_id}', "
            f"status={self.status.name}, "
            f"started_at={self.started_at}, "
            f"conclusion={self.conclusion}, "
            f"completed_at={self.completed_at}, "
            f"output={self.output}, "
            f"app={self.app})"
        )

    @property
    def name(self) -> str:
        """Name of the check run."""
        return self.raw_check_run.name

    @name.setter
    def name(self, name: str) -> None:
        self.raw_check_run.edit(name=name)

    @property
    def commit_sha(self) -> str:
        """Commit SHA that check run is related to."""
        return self.raw_check_run.head_sha

    @property
    def url(self) -> Optional[str]:
        """URL with additional details."""
        return self.raw_check_run.details_url

    @url.setter
    def url(self, url: str) -> None:
        self.raw_check_run.edit(details_url=url)

    @property
    def external_id(self) -> Optional[str]:
        """External ID that can be used internally by the integrator."""
        return self.raw_check_run.external_id

    @external_id.setter
    def external_id(self, external_id: str) -> None:
        self.raw_check_run.edit(external_id=external_id)

    @property
    def status(self) -> GithubCheckRunStatus:
        """Current status of the check run."""
        return GithubCheckRunStatus(self.raw_check_run.status)

    @property
    def started_at(self) -> Optional[datetime.datetime]:
        """Timestamp of start of the check run."""
        return self.raw_check_run.started_at

    @started_at.setter
    def started_at(self, started_at: datetime.datetime) -> None:
        self.raw_check_run.edit(started_at=started_at)

    @property
    def conclusion(self) -> Optional[GithubCheckRunResult]:
        """Conclusion/result of the check run."""
        return (
            GithubCheckRunResult(self.raw_check_run.conclusion)
            if self.raw_check_run.conclusion
            else None
        )

    @property
    def completed_at(self) -> Optional[datetime.datetime]:
        """Timestamp of completion of the check run."""
        return self.raw_check_run.completed_at

    @property
    def output(self) -> CheckRunOutput:
        """Output of the check run."""
        return self.raw_check_run.output

    @output.setter
    def output(self, output: GithubCheckRunOutput) -> None:
        self.raw_check_run.edit(output=output)

    @property
    def app(self) -> GithubApp:
        """Github App of the check run."""
        return self.raw_check_run.app

    def change_status(
        self,
        status: Optional[GithubCheckRunStatus] = None,
        completed_at: Optional[datetime.datetime] = None,
        conclusion: Optional[GithubCheckRunResult] = None,
    ) -> None:
        """
        Changes the status of the check run and checks the validity of new state.

        Args:
            status: Status of the check run to be set. If set to completed, you
                must provide conclusion.

                Defaults to `None`.
            completed_at: Timestamp of completion of the check run. If set, you
                must provide conclusion.

                Defaults to `None`.
            conclusion: Conclusion/result of the check run. If only conclusion
                is set, status is automatically set to completed.

                Defaults to `None`.

        Raises:
            OperationNotSupported, if given completed or timestamp of completed
                without conclusion.
        """
        if not (status or completed_at or conclusion):
            return

        if (
            status == GithubCheckRunStatus.completed or completed_at
        ) and conclusion is None:
            raise OperationNotSupported(
                "When provided completed status or completed at,"
                " you need to provide conclusion.",
            )

        self.raw_check_run.edit(
            status=value_or_NotSet(status.name if status else None),
            conclusion=value_or_NotSet(conclusion.name if conclusion else None),
            completed_at=value_or_NotSet(completed_at),
        )

    @staticmethod
    def get_list(
        project: "ogr_github.GithubProject",
        commit_sha: str,
        name: Optional[str] = None,
        status: Optional[GithubCheckRunStatus] = None,
    ) -> list["GithubCheckRun"]:
        """
        Returns list of GitHub check runs.

        Args:
            project: Project from which the check runs are retrieved.
            commit_sha: Commit to which are the check runs related to.
            name: Name of the check run for filtering.

                Defaults to `None`, no filtering.
            status: Status of the check runs to be returned.

                Defaults to `None`, no filtering.

        Returns:
            List of the check runs.
        """
        check_runs = project.github_repo.get_commit(commit_sha).get_check_runs(
            check_name=value_or_NotSet(name),
            status=value_or_NotSet(status.name if status else None),
        )

        return [GithubCheckRun(project, run) for run in check_runs]

    @staticmethod
    def get(
        project: "ogr_github.GithubProject",
        check_run_id: Optional[int] = None,
        commit_sha: Optional[str] = None,
    ) -> Optional["GithubCheckRun"]:
        """
        Retrieves GitHub check run as ogr object.

        Args:
            project: Project from which the check run is retrieved.
            check_run_id: Check run ID.

                Defaults to `None`, i.e. is not used for query.
            commit_sha: Commit SHA from which the check run is to be retrieved.
                If set, returns latest check run for the commit.

                Defaults to `None`, i.e. is not used for query.

        Returns:
            GithubCheckRun object or `None` if no check run is found.

        Raises:
            OperationNotSupported, in case there is no parameter for query set
                or both are set.
        """
        if check_run_id is not None and commit_sha:
            raise OperationNotSupported(
                "Cannot retrieve check run by both ID and commit hash",
            )

        if not (check_run_id is not None or commit_sha):
            raise OperationNotSupported("Cannot retrieve check run by no criteria")

        if check_run_id is not None:
            return GithubCheckRun(
                project,
                project.github_repo.get_check_run(check_run_id),
            )

        check_runs = project.github_repo.get_commit(commit_sha).get_check_runs()
        if check_runs.totalCount == 0:
            return None
        return GithubCheckRun(project, check_runs[0])

    @staticmethod
    def create(
        project: "ogr_github.GithubProject",
        name: str,
        commit_sha: str,
        url: Optional[str] = None,
        external_id: Optional[str] = None,
        status: GithubCheckRunStatus = GithubCheckRunStatus.queued,
        started_at: Optional[datetime.datetime] = None,
        conclusion: Optional[GithubCheckRunResult] = None,
        completed_at: Optional[datetime.datetime] = None,
        output: Optional[GithubCheckRunOutput] = None,
        actions: Optional[list[dict[str, str]]] = None,
    ) -> "GithubCheckRun":
        """
        Creates new check run.

        Args:
            project: Project where the check run is to be created.
            name: Name of the check run.
            commit_sha: Hash of the commit that check run is related to.
            url: URL with details of the run.

                Defaults to `None`.
            external_id: External ID that can be used internally by integrator.

                Defaults to `None`.
            status: Status of the check run.

                Defaults to queued.
            started_at: Timestamp of starting the check run.

                Defaults to `None`.
            conclusion: Conclusion of the check run. Should be set with status
                completed.

                Defaults to `None`.
            completed_at: Timestamp of completion of the check run. If set, you
                must provide conclusion.

                Defaults to `None`.
            output: Output of the check run.
            actions: List of possible follow-up actions for the check run.

        Returns:
            Created check run object.

        Raises:
            OperationNotSupported, if given completed status or completion
                timestamp and no conclusion.
        """

        if (
            completed_at or status == GithubCheckRunStatus.completed
        ) and conclusion is None:
            raise OperationNotSupported(
                "When provided completed_at or completed status, "
                "you need to provide conclusion.",
            )

        created_check_run = project.github_repo.create_check_run(
            name=name,
            head_sha=commit_sha,
            details_url=value_or_NotSet(url),
            external_id=value_or_NotSet(external_id),
            status=status.name,
            started_at=value_or_NotSet(started_at),
            conclusion=value_or_NotSet(conclusion.name if conclusion else None),
            completed_at=value_or_NotSet(completed_at),
            output=value_or_NotSet(output),
            actions=value_or_NotSet(actions),
        )

        return GithubCheckRun(project, created_check_run)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from requre.online_replacing import record_requests_for_all_methods

from ogr.abstract import IssueStatus
from ogr.exceptions import IssueTrackerDisabled
from tests.integration.pagure.base import PagureTests


@record_requests_for_all_methods()
class Issues(PagureTests):
    def setUp(self):
        super().setUp()
        self._long_issues_project = None

    @property
    def long_issues_project(self):
        if not self._long_issues_project:
            self._long_issues_project = self.service.get_project(
                repo="pagure",
                namespace=None,
            )

        return self._long_issues_project

    def test_issue_list(self):
        issue_list = self.ogr_project.get_issue_list()
        assert isinstance(issue_list, list)

        issue_list = self.ogr_project.get_issue_list(status=IssueStatus.all)
        assert issue_list
        assert len(issue_list) >= 2

    def test_issue_list_paginated(self):
        issue_list = self.long_issues_project.get_issue_list()
        assert issue_list
        assert len(issue_list) >= 400

    def test_issue_list_author(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            author="mfocko",
        )
        assert issue_list
        assert len(issue_list) >= 3

    def test_issue_list_nonexisting_author(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            author="xyzidontexist",
        )
        assert len(issue_list) == 0

    def test_issue_list_assignee(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            assignee="mfocko",
        )
        assert issue_list
        assert len(issue_list) == 1

    def test_issue_list_labels(self):
        issue_list = self.ogr_project.get_issue_list(
            status=IssueStatus.all,
            labels=["test_label"],
        )
        assert issue_list
        assert len(issue_list) == 1

    def test_create_issue(self):
        title = "This is an issue"
        description = "Example of Issue description"
        labels = ["label1", "label2"]
        project = self.service.get_project(repo="hello-112111", namespace="testing")
        issue = project.create_issue(
            title=title,
            body=description,
            private=True,
            labels=labels,
        )
        assert issue.title == title
        assert issue.description == description
        assert issue.private
        for issue_label, label in zip(issue.labels, labels):
            assert issue_label.name == label

    def test_create_issue_with_assignees(self):
        random_str = "something"
        project = self.service.get_project(repo="hello-112111", namespace="testing")
        assignee = ["mfocko"]
        issue = project.create_issue(
            title=random_str,
            body=random_str,
            assignees=assignee,
        )
        assert issue.title == random_str
        assert issue.description == random_str
        assert issue.assignee == assignee[0]

    def test_issue_assignees(self):
        """
        Remove the assignees from this issue before regenerating the response files:
        https://pagure.io/testing/hello-112111/issue/4
        """

        project = self.service.get_project(
            repo="hello-112111",
            namespace="testing",
            is_fork=True,
        )
        issue = project.get_issue(4)

        assert not project.get_issue(4).assignee
        issue.add_assignee("kpostlet")
        assignee = project.get_issue(4).assignee
        assert assignee == "kpostlet"

    def test_issue_without_label(self):
        title = "This is an issue"
        description = "Example of Issue description"
        project = self.service.get_project(repo="hello-112111", namespace="testing")
        issue = project.create_issue(title=title, body=description)
        assert issue.title == title
        assert issue.description == description

    def test_get_comment(self):
        project = self.service.get_project(
            repo="my-playground",
            namespace=None,
            username="nikromen",
        )
        comment = project.get_issue(1).get_comment(753462)
        assert comment.body == "example issue comment"

    def test_create_with_disabled_issues(self):
        with pytest.raises(IssueTrackerDisabled):
            self.ogr_project.get_fork().create_issue(
                "Testing issue",
                "shouldn't be created",
            )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from ogr.services.github.check_run import GithubCheckRun
from ogr.services.github.comments import GithubIssueComment, GithubPRComment
from ogr.services.github.issue import GithubIssue
from ogr.services.github.project import GithubProject
from ogr.services.github.pull_request import GithubPullRequest
from ogr.services.github.release import GithubRelease
from ogr.services.github.service import GithubService
from ogr.services.github.user import GithubUser

__all__ = [
    GithubCheckRun.__name__,
    GithubPullRequest.__name__,
    GithubIssueComment.__name__,
    GithubPRComment.__name__,
    GithubIssue.__name__,
    GithubRelease.__name__,
    GithubUser.__name__,
    GithubProject.__name__,
    GithubService.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from deprecated import deprecated


def deprecate_and_set_removal(since: str, remove_in: str, message: str):
    """
    Decorator for deprecating functions in ogr.

    Args:
        since: Indicates a version since which is attribute deprecated.
        remove_in: Indicates a version in which the attribute will be removed.
        message: Message to be included with deprecation.

    Returns:
        Decorator.
    """
    return deprecated(
        version=since,
        reason=f"will be removed in {remove_in}: {message}",
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import Union

from github.CommitComment import CommitComment as _GithubCommitComment
from github.IssueComment import IssueComment as _GithubIssueComment
from github.PullRequestComment import PullRequestComment as _GithubPullRequestComment
from github.Reaction import Reaction as _Reaction

from ogr.abstract import Comment, CommitComment, IssueComment, PRComment, Reaction


class GithubReaction(Reaction):
    _raw_reaction: _Reaction

    def __str__(self) -> str:
        return "Github" + super().__str__()

    def delete(self) -> None:
        self._raw_reaction.delete()


class GithubComment(Comment):
    def _from_raw_comment(
        self,
        raw_comment: Union[
            _GithubIssueComment,
            _GithubPullRequestComment,
            _GithubCommitComment,
        ],
    ) -> None:
        self._raw_comment = raw_comment
        self._id = raw_comment.id
        self._author = raw_comment.user.login
        self._created = raw_comment.created_at

    @property
    def body(self) -> str:
        return self._raw_comment.body

    @body.setter
    def body(self, new_body: str) -> None:
        self._raw_comment.edit(new_body)

    @property
    def edited(self) -> datetime.datetime:
        return self._raw_comment.updated_at

    def get_reactions(self) -> list[Reaction]:
        return [
            GithubReaction(reaction) for reaction in self._raw_comment.get_reactions()
        ]

    def add_reaction(self, reaction: str) -> GithubReaction:
        return GithubReaction(self._raw_comment.create_reaction(reaction))


class GithubIssueComment(GithubComment, IssueComment):
    def __str__(self) -> str:
        return "Github" + super().__str__()


class GithubPRComment(GithubComment, PRComment):
    def __str__(self) -> str:
        return "Github" + super().__str__()


class GithubCommitComment(GithubComment, CommitComment):
    def __str__(self) -> str:
        return "Github" + super().__str__()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from typing import Union

from github.Label import Label as _GithubLabel

from ogr.abstract import Issue, IssueLabel, Label, PRLabel, PullRequest


class GithubLabel(Label):
    def __init__(
        self,
        raw_label: _GithubLabel,
        parent: Union[PullRequest, Issue],
    ) -> None:
        super().__init__(parent)
        self._raw_label = raw_label

    def __str__(self) -> str:
        return f'GithubLabel(name="{self.name}")'

    @property
    def name(self):
        return self._raw_label.name


class GithubPRLabel(GithubLabel, PRLabel):
    pass


class GithubIssueLabel(GithubLabel, IssueLabel):
    pass



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import ClassVar

from github import UnknownObjectException

from ogr.abstract import CommitFlag, CommitStatus
from ogr.services import github as ogr_github
from ogr.services.base import BaseCommitFlag


class GithubCommitFlag(BaseCommitFlag):
    _states: ClassVar[dict[str, CommitStatus]] = {
        "pending": CommitStatus.pending,
        "success": CommitStatus.success,
        "failure": CommitStatus.failure,
        "error": CommitStatus.error,
    }

    def __str__(self) -> str:
        return "Github" + super().__str__()

    def _from_raw_commit_flag(self):
        self.state = self._state_from_str(self._raw_commit_flag.state)
        self.context = self._raw_commit_flag.context
        self.comment = self._raw_commit_flag.description
        self.url = self._raw_commit_flag.target_url
        self.uid = self._raw_commit_flag.id

    @staticmethod
    def get(project: "ogr_github.GithubProject", commit: str) -> list["CommitFlag"]:
        statuses = project.github_repo.get_commit(commit).get_statuses()

        try:
            return [
                GithubCommitFlag(
                    raw_commit_flag=raw_status,
                    project=project,
                    commit=commit,
                )
                for raw_status in statuses
            ]
        except UnknownObjectException:
            return []

    @staticmethod
    def set(
        project: "ogr_github.GithubProject",
        commit: str,
        state: CommitStatus,
        target_url: str,
        description: str,
        context: str,
        trim: bool = False,
    ) -> "CommitFlag":
        state = GithubCommitFlag._validate_state(state)

        github_commit = project.github_repo.get_commit(commit)
        if trim:
            description = description[:140]
        status = github_commit.create_status(
            state.name,
            target_url,
            description,
            context,
        )
        return GithubCommitFlag(project=project, raw_commit_flag=status, commit=commit)

    @property
    def created(self) -> datetime.datetime:
        return self._raw_commit_flag.created_at

    @property
    def edited(self) -> datetime.datetime:
        return self._raw_commit_flag.updated_at



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
from typing import ClassVar, Optional

import gitlab
import requests
from gitlab.exceptions import GitlabGetError
from gitlab.v4.objects import MergeRequest as _GitlabMergeRequest

from ogr.abstract import MergeCommitStatus, PRComment, PRLabel, PRStatus, PullRequest
from ogr.exceptions import GitlabAPIException, OgrNetworkError
from ogr.services import gitlab as ogr_gitlab
from ogr.services.base import BasePullRequest
from ogr.services.gitlab.comments import GitlabPRComment
from ogr.services.gitlab.label import GitlabPRLabel


class GitlabPullRequest(BasePullRequest):
    _raw_pr: _GitlabMergeRequest
    _target_project: "ogr_gitlab.GitlabProject"
    _source_project: "ogr_gitlab.GitlabProject" = None
    _merge_commit_status: ClassVar[dict[str, MergeCommitStatus]] = {
        "can_be_merged": MergeCommitStatus.can_be_merged,
        "cannot_be_merged": MergeCommitStatus.cannot_be_merged,
        "unchecked": MergeCommitStatus.unchecked,
        "checking": MergeCommitStatus.checking,
        "cannot_be_merged_recheck": MergeCommitStatus.cannot_be_merged_recheck,
    }

    @property
    def title(self) -> str:
        return self._raw_pr.title

    @title.setter
    def title(self, new_title: str) -> None:
        self._raw_pr.title = new_title
        self._raw_pr.save()

    @property
    def id(self) -> int:
        return self._raw_pr.iid

    @property
    def status(self) -> PRStatus:
        return (
            PRStatus.open
            if self._raw_pr.state == "opened"
            else PRStatus[self._raw_pr.state]
        )

    @property
    def url(self) -> str:
        return self._raw_pr.web_url

    @property
    def description(self) -> str:
        return self._raw_pr.description

    @description.setter
    def description(self, new_description: str) -> None:
        self._raw_pr.description = new_description
        self._raw_pr.save()

    @property
    def author(self) -> str:
        return self._raw_pr.author["username"]

    @property
    def source_branch(self) -> str:
        return self._raw_pr.source_branch

    @property
    def target_branch(self) -> str:
        return self._raw_pr.target_branch

    @property
    def created(self) -> datetime.datetime:
        return self._raw_pr.created_at

    @property
    def labels(self) -> list[PRLabel]:
        return [GitlabPRLabel(label, self) for label in self._raw_pr.labels]

    @property
    def diff_url(self) -> str:
        return f"{self._raw_pr.web_url}/diffs"

    @property
    def commits_url(self) -> str:
        return f"{self._raw_pr.web_url}/commits"

    @property
    def patch(self) -> bytes:
        response = requests.get(f"{self.url}.patch")

        if not response.ok:
            cls = OgrNetworkError if response.status_code >= 500 else GitlabAPIException
            raise cls(
                f"Couldn't get patch from {self.url}.patch because {response.reason}.",
            )

        return response.content

    @property
    def head_commit(self) -> str:
        return self._raw_pr.sha

    @property
    def merge_commit_sha(self) -> Optional[str]:
        # when merged => return merge_commit_sha
        # otherwise => return test merge if possible
        if self.status == PRStatus.merged:
            return self._raw_pr.merge_commit_sha

        # works for test merge only with python-gitlab>=2.10.0
        try:
            response = self._raw_pr.merge_ref()
        except GitlabGetError as ex:
            if ex.response_code == 400:
                return None
            raise
        return response.get("commit_id")

    @property
    def merge_commit_status(self) -> MergeCommitStatus:
        status = self._raw_pr.merge_status
        if status not in self._merge_commit_status:
            raise GitlabAPIException(f"Invalid merge_status {status}")

        return self._merge_commit_status[status]

    @property
    def source_project(self) -> "ogr_gitlab.GitlabProject":
        if self._source_project is None:
            self._source_project = (
                self._target_project.service.get_project_from_project_id(
                    self._raw_pr.attributes["source_project_id"],
                )
            )
        return self._source_project

    def __str__(self) -> str:
        return "Gitlab" + super().__str__()

    @staticmethod
    def create(
        project: "ogr_gitlab.GitlabProject",
        title: str,
        body: str,
        target_branch: str,
        source_branch: str,
        fork_username: Optional[str] = None,
    ) -> "PullRequest":
        """
        How to create PR:
        -  upstream -> upstream - call on upstream, fork_username unset
        -  fork -> upstream - call on fork, fork_username unset
           also can call on upstream with fork_username, not supported way of using
        -  fork -> fork - call on fork, fork_username set
        -  fork -> other_fork - call on fork, fork_username set to other_fork owner
        """
        repo = project.gitlab_repo
        parameters = {
            "source_branch": source_branch,
            "target_branch": target_branch,
            "title": title,
            "description": body,
        }
        target_id = None

        target_project = project
        if project.is_fork and fork_username is None:
            # handles fork -> upstream (called on fork)
            target_id = project.parent.gitlab_repo.attributes["id"]
            target_project = project.parent
        elif fork_username and fork_username != project.namespace:
            # handles fork -> upstream
            #   (username of fork owner specified by fork_username)
            # handles fork -> other_fork
            #   (username of other_fork owner specified by fork_username)

            other_project = GitlabPullRequest.__get_fork(
                fork_username,
                project if project.parent is None else project.parent,
            )

            target_id = other_project.gitlab_repo.attributes["id"]
            if project.parent is None:
                target_id = repo.attributes["id"]
                repo = other_project.gitlab_repo
        # otherwise handles PR from the same project to same project

        if target_id is not None:
            parameters["target_project_id"] = target_id

        mr = repo.mergerequests.create(parameters)
        return GitlabPullRequest(mr, target_project)

    @staticmethod
    def __get_fork(
        fork_username: str,
        project: "ogr_gitlab.GitlabProject",
    ) -> "ogr_gitlab.GitlabProject":
        """
        Returns forked project of a requested user. Internal method, in case the fork
        doesn't exist, raises GitlabAPIException.

        Args:
            fork_username: Username of a user that owns requested fork.
            project: Project to search forks of.

        Returns:
            Requested fork.

        Raises:
            GitlabAPIException, in case the fork doesn't exist.
        """
        forks = list(
            filter(
                lambda fork: fork.gitlab_repo.namespace["full_path"] == fork_username,
                project.get_forks(),
            ),
        )
        if not forks:
            raise GitlabAPIException("Requested fork doesn't exist")
        return forks[0]

    @staticmethod
    def get(project: "ogr_gitlab.GitlabProject", pr_id: int) -> "PullRequest":
        try:
            mr = project.gitlab_repo.mergerequests.get(pr_id)
        except gitlab.GitlabGetError as ex:
            raise GitlabAPIException(f"No PR with id {pr_id} found") from ex
        return GitlabPullRequest(mr, project)

    @staticmethod
    def get_list(
        project: "ogr_gitlab.GitlabProject",
        status: PRStatus = PRStatus.open,
    ) -> list["PullRequest"]:
        # Gitlab API has status 'opened', not 'open'
        # f"Calling a `list()` method without specifying `get_all=True` or "
        # f"`iterator=True` will return a maximum of 20 items. "
        mrs = project.gitlab_repo.mergerequests.list(
            state=status.name if status != PRStatus.open else "opened",
            order_by="updated_at",
            sort="desc",
            # gitlab 3.3 syntax was all=True
            get_all=True,
        )
        return [GitlabPullRequest(mr, project) for mr in mrs]

    def update_info(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "PullRequest":
        if title:
            self._raw_pr.title = title
        if description:
            self._raw_pr.description = description

        self._raw_pr.save()
        return self

    def _get_all_comments(self) -> list[PRComment]:
        return [
            GitlabPRComment(parent=self, raw_comment=raw_comment)
            for raw_comment in self._raw_pr.notes.list(sort="asc", all=True)
        ]

    def get_all_commits(self) -> list[str]:
        return [commit.id for commit in self._raw_pr.commits()]

    def comment(
        self,
        body: str,
        commit: Optional[str] = None,
        filename: Optional[str] = None,
        row: Optional[int] = None,
    ) -> "PRComment":
        comment = self._raw_pr.notes.create({"body": body})
        return GitlabPRComment(parent=self, raw_comment=comment)

    def close(self) -> "PullRequest":
        self._raw_pr.state_event = "close"
        self._raw_pr.save()
        return self

    def merge(self) -> "PullRequest":
        self._raw_pr.merge()
        return self

    def add_label(self, *labels: str) -> None:
        self._raw_pr.labels += labels
        self._raw_pr.save()

    def get_comment(self, comment_id: int) -> PRComment:
        return GitlabPRComment(self._raw_pr.notes.get(comment_id))



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Module providing one api for multiple git services (github/gitlab/pagure)
"""

import contextlib
from importlib.metadata import PackageNotFoundError, distribution

from ogr.abstract import AuthMethod
from ogr.factory import (
    get_instances_from_dict,
    get_project,
    get_service_class,
    get_service_class_or_none,
)
from ogr.services.github import GithubService
from ogr.services.gitlab import GitlabService
from ogr.services.pagure import PagureService

with contextlib.suppress(PackageNotFoundError):
    __version__ = distribution(__name__).version

__all__ = [
    GithubService.__name__,
    PagureService.__name__,
    GitlabService.__name__,
    AuthMethod.__name__,
    get_project.__name__,
    get_service_class.__name__,
    get_service_class_or_none.__name__,
    get_instances_from_dict.__name__,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import logging
from time import sleep
from typing import Any, Optional, Union

from ogr.abstract import (
    CommitFlag,
    CommitStatus,
    PRComment,
    PRLabel,
    PRStatus,
    PullRequest,
)
from ogr.exceptions import PagureAPIException
from ogr.services import pagure as ogr_pagure
from ogr.services.base import BasePullRequest
from ogr.services.pagure.comments import PagurePRComment
from ogr.services.pagure.label import PagurePRLabel

logger = logging.getLogger(__name__)


class PagurePullRequest(BasePullRequest):
    _target_project: "ogr_pagure.PagureProject"
    _source_project: "ogr_pagure.PagureProject" = None

    def __init__(self, raw_pr, project):
        super().__init__(raw_pr, project)
        self.__dirty = False

    def __update(self):
        if self.__dirty:
            self._raw_pr = self.__call_api()
            self.__dirty = False

    @property
    def title(self) -> str:
        self.__update()
        return self._raw_pr["title"]

    @title.setter
    def title(self, new_title: str) -> None:
        self.update_info(title=new_title)

    @property
    def id(self) -> int:
        return self._raw_pr["id"]

    @property
    def status(self) -> PRStatus:
        self.__update()
        return PRStatus[self._raw_pr["status"].lower()]

    @property
    def url(self) -> str:
        return "/".join(
            [
                self.target_project.service.instance_url,
                self._raw_pr["project"]["url_path"],
                "pull-request",
                str(self.id),
            ],
        )

    @property
    def description(self) -> str:
        self.__update()
        return self._raw_pr["initial_comment"]

    @description.setter
    def description(self, new_description: str) -> None:
        self.update_info(description=new_description)

    @property
    def author(self) -> str:
        return self._raw_pr["user"]["name"]

    @property
    def source_branch(self) -> str:
        return self._raw_pr["branch_from"]

    @property
    def target_branch(self) -> str:
        return self._raw_pr["branch"]

    @property
    def created(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(int(self._raw_pr["date_created"]))

    @property
    def diff_url(self) -> str:
        return f"{self.url}#request_diff"

    @property
    def commits_url(self) -> str:
        return f"{self.url}#commit_list"

    @property
    def patch(self) -> bytes:
        request_response = self._target_project._call_project_api_raw(
            "pull-request",
            f"{self.id}.patch",
            add_api_endpoint_part=False,
        )
        if request_response.status_code != 200:
            raise PagureAPIException(
                f"Cannot get patch from {self.url}.patch because {request_response.reason}.",
                response_code=request_response.status_code,
            )
        return request_response.content

    @property
    def head_commit(self) -> str:
        return self._raw_pr["commit_stop"]

    @property
    def source_project(self) -> "ogr_pagure.PagureProject":
        if self._source_project is None:
            source = self._raw_pr["repo_from"]
            source_project_info = {
                "repo": source["name"],
                "namespace": source["namespace"],
            }

            if source["parent"] is not None:
                source_project_info["is_fork"] = True
                source_project_info["username"] = source["user"]["name"]

            self._source_project = self._target_project.service.get_project(
                **source_project_info,
            )

        return self._source_project

    @property
    def closed_by(self) -> Optional[str]:
        closed_by = self._raw_pr["closed_by"]
        return closed_by["name"] if closed_by else None

    @property
    def labels(self) -> list[PRLabel]:
        return [PagurePRLabel(label, self) for label in self._raw_pr["tags"]]

    def __str__(self) -> str:
        return "Pagure" + super().__str__()

    def __call_api(self, *args, **kwargs) -> dict:
        return self._target_project._call_project_api(
            "pull-request",
            str(self.id),
            *args,
            **kwargs,
        )

    @staticmethod
    def create(
        project: "ogr_pagure.PagureProject",
        title: str,
        body: str,
        target_branch: str,
        source_branch: str,
        fork_username: Optional[str] = None,
    ) -> "PullRequest":
        data = {
            "title": title,
            "branch_to": target_branch,
            "branch_from": source_branch,
            "initial_comment": body,
        }

        caller = project
        if project.is_fork:
            data["repo_from"] = project.repo
            data["repo_from_username"] = project._user
            data["repo_from_namespace"] = project.namespace

            # running the call from the parent project
            caller = caller.parent
        elif fork_username:
            fork_project = project.service.get_project(
                username=fork_username,
                repo=project.repo,
                namespace=project.namespace,
                is_fork=True,
            )
            data["repo_from_username"] = fork_username
            data["repo_from"] = fork_project.repo
            data["repo_from_namespace"] = fork_project.namespace

        response = caller._call_project_api(
            "pull-request",
            "new",
            method="POST",
            data=data,
        )
        return PagurePullRequest(response, caller)

    @staticmethod
    def get(project: "ogr_pagure.PagureProject", pr_id: int) -> "PullRequest":
        raw_pr = project._call_project_api("pull-request", str(pr_id))
        return PagurePullRequest(raw_pr, project)

    @staticmethod
    def get_files_diff(
        project: "ogr_pagure.PagureProject",
        pr_id: int,
        retries: int = 0,
        wait_seconds: int = 3,
    ) -> dict:
        """
        Retrieve pull request diff statistics.

        Pagure API tends to return ENOPRSTATS error when a pull request is transitioning
        from open to other states, so you can use `retries` and `wait_seconds` to try to
        mitigate that.


        Args:
            project: Pagure project.
            pr_id: Pull request ID.
            retries: Number of extra attempts.
            wait_seconds: Delay between attempts.
        """
        attempt = 1
        while True:
            try:
                return project._call_project_api(
                    "pull-request",
                    str(pr_id),
                    "diffstats",
                    method="GET",
                )
            except PagureAPIException as ex:  # noqa PERF203
                if "No statistics" in ex.pagure_error:
                    # this may be a race condition, try once more
                    logger.info(
                        f"While retrieving PR diffstats Pagure returned ENOPRSTATS.\n{ex}",
                    )
                    if attempt <= retries:
                        attempt += 1
                        logger.info(
                            f"Trying again; attempt={attempt} after {wait_seconds} seconds",
                        )
                        sleep(wait_seconds)
                        continue
                raise ex

    @staticmethod
    def get_list(
        project: "ogr_pagure.PagureProject",
        status: PRStatus = PRStatus.open,
        assignee=None,
        author=None,
    ) -> list["PullRequest"]:
        payload = {"page": 1, "status": status.name.capitalize()}
        if assignee is not None:
            payload["assignee"] = assignee
        if author is not None:
            payload["author"] = author

        raw_prs = []
        while True:
            page_result = project._call_project_api("pull-requests", params=payload)
            raw_prs += page_result["requests"]
            if not page_result["pagination"]["next"]:
                break

            # mypy don't know that key "page" really contains int...
            payload["page"] += 1  # type: ignore

        return [PagurePullRequest(pr_dict, project) for pr_dict in raw_prs]

    def update_info(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "PullRequest":
        try:
            data = {"title": title if title else self.title}

            if description:
                data["initial_comment"] = description

            updated_pr = self.__call_api(method="POST", data=data)
            logger.info("PR updated.")

            self._raw_pr = updated_pr
            return self
        except Exception as ex:
            raise PagureAPIException("there was an error while updating the PR") from ex

    def _get_all_comments(self) -> list[PRComment]:
        self.__update()
        raw_comments = self._raw_pr["comments"]
        return [
            PagurePRComment(parent=self, raw_comment=comment_dict)
            for comment_dict in raw_comments
        ]

    def comment(
        self,
        body: str,
        commit: Optional[str] = None,
        filename: Optional[str] = None,
        row: Optional[int] = None,
    ) -> "PRComment":
        payload: dict[str, Any] = {"comment": body}
        if commit is not None:
            payload["commit"] = commit
        if filename is not None:
            payload["filename"] = filename
        if row is not None:
            payload["row"] = row

        self.__call_api("comment", method="POST", data=payload)
        self.__dirty = True
        return PagurePRComment(
            parent=self,
            body=body,
            author=self.target_project.service.user.get_username(),
        )

    def close(self) -> "PullRequest":
        return_value = self.__call_api("close", method="POST")

        if return_value["message"] != "Pull-request closed!":
            raise PagureAPIException(return_value["message"])

        self.__dirty = True
        return self

    def merge(self) -> "PullRequest":
        return_value = self.__call_api("merge", method="POST")

        if return_value["message"] != "Changes merged!":
            raise PagureAPIException(return_value["message"])

        self.__dirty = True
        return self

    def get_statuses(self) -> list[CommitFlag]:
        self.__update()
        return self.target_project.get_commit_statuses(self._raw_pr["commit_stop"])

    def set_flag(
        self,
        username: str,
        comment: str,
        url: str,
        status: Optional[CommitStatus] = None,
        percent: Optional[int] = None,
        uid: Optional[str] = None,
    ) -> dict:
        """
        Set a flag on a pull-request to display results or status of CI tasks.

        See "Flag a pull-request" at https://pagure.io/api/0/#pull_requests-tab
        for a full description of the parameters.

        Args:
            username: The name of the application to be presented to users
                on the pull request page.
            comment: A short message summarizing the presented results.
            url: A URL to the result of this flag.
            status: The status to be displayed for this flag.
            percent: A percentage of completion compared to the goal.
            uid: A unique identifier used to identify a flag on the pull-request.

        Returns:
            Dictionary with the response received from Pagure.
        """
        data: dict[str, Union[str, int]] = {
            "username": username,
            "comment": comment,
            "url": url,
        }
        if status is not None:
            data["status"] = status.name
        if percent is not None:
            data["percent"] = percent
        if uid is not None:
            data["uid"] = uid
        return self.__call_api("flag", method="POST", data=data)

    def get_comment(self, comment_id: int) -> PRComment:
        for comment in self._get_all_comments():
            if comment.id == comment_id:
                return comment

        raise PagureAPIException(
            f"No comment with id#{comment_id} in PR#{self.id} found.",
            response_code=404,
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from requre.online_replacing import record_requests_for_all_methods

from ogr.exceptions import OperationNotSupported, PagureAPIException
from tests.integration.pagure.base import PagureTests


@record_requests_for_all_methods()
class Forks(PagureTests):
    def test_fork(self):
        assert self.ogr_fork.exists()
        assert self.ogr_fork.is_fork
        fork_description = self.ogr_fork.get_description()
        assert fork_description
        a = self.ogr_fork.parent
        assert a
        is_forked = a.is_forked()
        assert is_forked
        assert isinstance(is_forked, bool)
        fork = a.get_fork(create=False)
        assert fork
        assert fork.is_fork
        urls = fork.get_git_urls()
        assert "{username}" not in urls["ssh"]

    def test_fork_in_str(self):
        str_representation = str(self.ogr_fork)
        assert 'username="' in str_representation
        assert "is_fork=True" in str_representation

    def test_nonexisting_fork(self):
        ogr_project_non_existing_fork = self.service.get_project(
            namespace=None,
            repo="ogr-tests",
            username="qwertzuiopasdfghjkl",
            is_fork=True,
        )
        assert not ogr_project_non_existing_fork.exists()
        with pytest.raises(PagureAPIException) as ex:
            ogr_project_non_existing_fork.get_description()
        assert "Project not found" in ex.value.pagure_error

    def test_fork_property(self):
        fork = self.ogr_project.get_fork()
        assert fork
        assert fork.get_description()

    def test_create_fork(self):
        """
        Remove your fork of ogr-tests https://pagure.io/fork/$USER/ogr-tests
        before regeneration data.
        But other tests needs to have already existed user fork.
        So regenerate data for other tests, remove  data file for this test
        and regenerate it again.
        """
        not_existing_fork = self.ogr_project.get_fork(create=False)
        assert not not_existing_fork
        assert not self.ogr_project.is_forked()

        old_forks = self.ogr_project.service.user.get_forks()

        self.ogr_project.fork_create()

        assert self.ogr_project.get_fork().exists()
        assert self.ogr_project.is_forked()

        new_forks = self.ogr_project.service.user.get_forks()
        assert len(old_forks) == len(new_forks) - 1

    def test_create_fork_with_namespace(self):
        with pytest.raises(OperationNotSupported):
            self.ogr_project.fork_create(namespace="some_random_namespace")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import re
from typing import Optional, Union

import github
import github.GithubObject
from github import (
    Github as PyGithubInstance,
)
from github import (
    Repository as PyGithubRepository,
)
from github import (
    UnknownObjectException,
)
from urllib3.util import Retry

from ogr.abstract import AuthMethod, GitUser
from ogr.exceptions import GithubAPIException
from ogr.factory import use_for_service
from ogr.services.base import BaseGitService, GitProject
from ogr.services.github.auth_providers import (
    GithubApp,
    GithubAuthentication,
    TokenAuthentication,
    Tokman,
)
from ogr.services.github.project import GithubProject
from ogr.services.github.user import GithubUser

logger = logging.getLogger(__name__)


@use_for_service("github.com")
class GithubService(BaseGitService):
    # class parameter could be used to mock Github class api
    github_class: type[github.Github]
    instance_url = "https://github.com"

    def __init__(
        self,
        token=None,
        read_only=False,
        github_app_id: Optional[str] = None,
        github_app_private_key: Optional[str] = None,
        github_app_private_key_path: Optional[str] = None,
        tokman_instance_url: Optional[str] = None,
        github_authentication: GithubAuthentication = None,
        max_retries: Union[int, Retry] = 1,
        **kwargs,
    ):
        """
        If multiple authentication methods are provided, they are prioritised:
            1. Tokman
            2. GithubApp
            3. TokenAuthentication (which is also default one, that works without specified token)
        """
        super().__init__()
        self.read_only = read_only
        self._default_auth_method = github_authentication
        self._other_auth_method: GithubAuthentication = None
        self._auth_methods: dict[AuthMethod, GithubAuthentication] = {}

        if isinstance(max_retries, Retry):
            self._max_retries = max_retries
        else:
            self._max_retries = Retry(
                total=int(max_retries),
                # Retry mechanism active for these HTTP methods:
                allowed_methods=["DELETE", "GET", "PATCH", "POST", "PUT"],
                # Only retry on following HTTP status codes
                status_forcelist=[500, 503, 403, 401],
                raise_on_status=True,
            )

        if not self._default_auth_method:
            self.__set_authentication(
                token=token,
                github_app_id=github_app_id,
                github_app_private_key=github_app_private_key,
                github_app_private_key_path=github_app_private_key_path,
                tokman_instance_url=tokman_instance_url,
                max_retries=self._max_retries,
            )

        if kwargs:
            logger.warning(f"Ignored keyword arguments: {kwargs}")

    def __set_authentication(self, **kwargs):
        auth_methods = [
            (Tokman, AuthMethod.tokman),
            (GithubApp, AuthMethod.github_app),
            (TokenAuthentication, AuthMethod.token),
        ]
        for auth_class, auth_name in auth_methods:
            auth_inst = auth_class.try_create(**kwargs)
            self._auth_methods[auth_name] = auth_inst
            if not self._default_auth_method:
                self._default_auth_method = auth_inst

        return None if self._default_auth_method else TokenAuthentication(None)

    def set_auth_method(self, method: AuthMethod):
        if self._auth_methods[method]:
            logger.info("Forced Github auth method to %s", method)
            self._other_auth_method = self._auth_methods[method]
        else:
            raise GithubAPIException(
                f"Choosen authentication method ({method}) is not available",
            )

    def reset_auth_method(self):
        logger.info("Reset Github auth method to the default")
        self._other_auth_method = None

    @property
    def authentication(self):
        return self._other_auth_method or self._default_auth_method

    @property
    def github(self):
        return self.authentication.pygithub_instance

    def __str__(self) -> str:
        readonly_str = ", read_only=True" if self.read_only else ""
        arguments = f", github_authentication={self.authentication!s}{readonly_str}"

        if arguments:
            # remove the first '- '
            arguments = arguments[2:]

        return f"GithubService({arguments})"

    def __eq__(self, o: object) -> bool:
        if not issubclass(o.__class__, GithubService):
            return False

        return (
            self.read_only == o.read_only  # type: ignore
            and self.authentication == o.authentication  # type: ignore
        )

    def __hash__(self) -> int:
        return hash(str(self))

    def get_project(
        self,
        repo=None,
        namespace=None,
        is_fork=False,
        **kwargs,
    ) -> "GithubProject":
        if is_fork:
            namespace = self.user.get_username()
        return GithubProject(
            repo=repo,
            namespace=namespace,
            service=self,
            read_only=self.read_only,
            **kwargs,
        )

    def get_project_from_github_repository(
        self,
        github_repo: PyGithubRepository.Repository,
    ) -> "GithubProject":
        return GithubProject(
            repo=github_repo.name,
            namespace=github_repo.owner.login,
            github_repo=github_repo,
            service=self,
            read_only=self.read_only,
        )

    @property
    def user(self) -> GitUser:
        return GithubUser(service=self)

    def change_token(self, new_token: str) -> None:
        self._default_auth_method = TokenAuthentication(new_token)

    def project_create(
        self,
        repo: str,
        namespace: Optional[str] = None,
        description: Optional[str] = None,
    ) -> "GithubProject":
        if namespace:
            try:
                owner = self.github.get_organization(namespace)
            except UnknownObjectException as ex:
                raise GithubAPIException(f"Group {namespace} not found.") from ex
        else:
            owner = self.github.get_user()

        try:
            new_repo = owner.create_repo(
                name=repo,
                description=description if description else github.GithubObject.NotSet,
            )
        except github.GithubException as ex:
            raise GithubAPIException("Project creation failed") from ex
        return GithubProject(
            repo=repo,
            namespace=namespace or owner.login,
            service=self,
            github_repo=new_repo,
        )

    def get_pygithub_instance(self, namespace: str, repo: str) -> PyGithubInstance:
        token = self.authentication.get_token(namespace, repo)
        return PyGithubInstance(login_or_token=token, retry=self._max_retries)

    def list_projects(
        self,
        namespace: Optional[str] = None,
        user: Optional[str] = None,
        search_pattern: Optional[str] = None,
        language: Optional[str] = None,
    ) -> list[GitProject]:
        search_query = ""

        if user:
            search_query += f"user:{user}"

        if language:
            search_query += f" language:{language}"

        projects: list[GitProject]
        projects = [
            GithubProject(
                repo=repo.name,
                namespace=repo.owner.login,
                github_repo=repo,
                service=self,
            )
            for repo in self.github.search_repositories(search_query, order="asc")
        ]

        if search_pattern:
            projects = [
                project
                for project in projects
                if re.search(search_pattern, project.repo)
            ]

        return projects



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from datetime import datetime

from requre.online_replacing import record_requests_for_all_methods

from tests.integration.github.base import GithubTests


@record_requests_for_all_methods()
class Comments(GithubTests):
    def test_pr_comments(self):
        pr_comments = self.ogr_project.get_pr(9).get_comments()
        assert pr_comments
        assert len(pr_comments) == 2

        assert pr_comments[0].body.endswith("fixed")
        assert pr_comments[1].body.startswith("LGTM")

    def test_pr_comments_reversed(self):
        pr_comments = self.ogr_project.get_pr(9).get_comments(reverse=True)
        assert pr_comments
        assert len(pr_comments) == 2
        assert pr_comments[0].body.startswith("LGTM")

    def test_pr_comments_filter(self):
        pr_comments = self.ogr_project.get_pr(9).get_comments(filter_regex="fixed")
        assert pr_comments
        assert len(pr_comments) == 1
        assert pr_comments[0].body.startswith("@TomasTomecek")

        pr_comments = self.ogr_project.get_pr(9).get_comments(
            filter_regex="LGTM, nicely ([a-z]*)",
        )
        assert pr_comments
        assert len(pr_comments) == 1
        assert pr_comments[0].body.endswith("done!")

    def test_pr_comments_search(self):
        comment_match = self.ogr_project.get_pr(9).search(filter_regex="LGTM")
        assert comment_match
        assert comment_match[0] == "LGTM"

        comment_match = self.ogr_project.get_pr(9).search(
            filter_regex="LGTM, nicely ([a-z]*)",
        )
        assert comment_match
        assert comment_match[0] == "LGTM, nicely done"

    def test_issue_comments(self):
        comments = self.ogr_project.get_issue(194).get_comments()
        assert len(comments) == 6
        assert comments[0].body.startswith("/packit")

    def test_issue_comments_reversed(self):
        comments = self.ogr_project.get_issue(194).get_comments(reverse=True)
        assert len(comments) == 6
        assert comments[0].body.startswith("The ")

    def test_issue_comments_regex(self):
        comments = self.ogr_project.get_issue(194).get_comments(
            filter_regex=r".*Fedora package.*",
        )
        assert len(comments) == 3
        assert "master" in comments[0].body

    def test_issue_comments_regex_reversed(self):
        comments = self.ogr_project.get_issue(194).get_comments(
            reverse=True,
            filter_regex=".*Fedora package.*",
        )
        assert len(comments) == 3
        assert "f29" in comments[0].body

    def test_pr_comments_author_regex(self):
        comments = self.ogr_project.get_pr(217).get_comments(
            filter_regex="^I",
            author="mfocko",
        )
        assert len(comments) == 1
        assert "API" in comments[0].body

    def test_pr_comments_author(self):
        comments = self.ogr_project.get_pr(217).get_comments(author="lachmanfrantisek")
        assert len(comments) == 3
        assert comments[0].body.endswith("here.")

    def test_issue_comments_author_regex(self):
        comments = self.ogr_project.get_issue(220).get_comments(
            filter_regex=".*API.*",
            author="lachmanfrantisek",
        )
        assert len(comments) == 1
        assert comments[0].body.startswith("After")

    def test_issue_comments_author(self):
        comments = self.ogr_project.get_issue(220).get_comments(author="mfocko")
        assert len(comments) == 2
        assert comments[0].body.startswith("What")
        assert comments[1].body.startswith("Consider")

    def test_issue_comments_updates(self):
        comments = self.hello_world_project.get_issue(61).get_comments(
            filter_regex="comment-update",
        )
        assert len(comments) == 1
        before_comment = comments[0].body
        before_edited = comments[0].edited

        comments[0].body = "see if updating works"
        assert comments[0].body == "see if updating works"
        assert comments[0].edited > before_edited

        comments[0].body = before_comment
        assert comments[0].body == before_comment

    def test_pr_comments_updates(self):
        comments = self.hello_world_project.get_pr(72).get_comments(
            filter_regex="comment updates",
        )
        assert len(comments) == 1
        before_comment = comments[0].body
        before_edited = comments[0].edited

        comments[0].body = "see if updating works"
        assert comments[0].body == "see if updating works"
        assert comments[0].edited > before_edited

        comments[0].body = before_comment
        assert comments[0].body == before_comment

    def test_pr_react_to_comment_and_delete(self):
        pr = self.service.get_project(repo="playground", namespace="nikromen").get_pr(4)
        pr_comment = pr.comment(datetime.now().strftime("%m/%d/%Y"))

        reaction = pr_comment.add_reaction("+1")
        assert len(pr_comment.get_reactions()) == 1

        reaction.delete()
        assert len(pr_comment.get_reactions()) == 0

    def test_issue_react_to_comment_and_delete(self):
        issue = self.service.get_project(
            repo="playground",
            namespace="nikromen",
        ).get_issue(5)
        issue_comment = issue.comment(datetime.now().strftime("%m/%d/%Y"))

        reaction = issue_comment.add_reaction("confused")
        assert len(issue_comment.get_reactions()) == 1

        reaction.delete()
        assert len(issue_comment.get_reactions()) == 0

    def test_get_reactions(self):
        pr = self.service.get_project(repo="playground", namespace="nikromen").get_pr(4)
        pr_comment = pr.comment(datetime.now().strftime("%m/%d/%Y"))

        pr_comment.add_reaction("+1")
        pr_comment.add_reaction("-1")
        pr_comment.add_reaction("confused")
        assert len(pr_comment.get_reactions()) == 3

        issue = self.service.get_project(
            repo="playground",
            namespace="nikromen",
        ).get_issue(5)
        issue_comment = issue.comment(datetime.now().strftime("%m/%d/%Y"))

        issue_comment.add_reaction("+1")
        issue_comment.add_reaction("-1")
        issue_comment.add_reaction("confused")
        assert len(pr_comment.get_reactions()) == 3
